/* gcd__odd look out for x>m
   gcdext can write direct to w if RALLOC big enough */


/* Various generic modular routines.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "mpm.h"
#include "mpm-impl.h"


/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x) 


#if GMP_LIMB_BITS == 32
#define DEADBEEF   (CNST_LIMB(0xDEADBEEF) & GMP_NUMB_MASK)
#endif
#if GMP_LIMB_BITS == 64
#define DEADBEEF   (CNST_LIMB(0xDEADBEEFBAADCAFE) & GMP_NUMB_MASK)
#endif

#if WANT_ASSERT
#define ASSERT_MPR_GEN(r,m)                             \
  do {                                                  \
    ASSERT (MRSIZE(m) >= 1);                            \
    ASSERT_MPN (RPTR(r), MRSIZE(m));                    \
  } while (0)
#else
#define ASSERT_MPR_GEN(r,m)  do { } while (0)
#endif


void
mpr_gen__clear (mpr_ptr r, mpm_srcptr m)
{
  __GMP_FREE_FUNC_LIMBS (RPTR(r), MRALLOC(m));
  /* ASSERT_CODE (RPTR(r) = NULL); */
}

/* using invert and mul, any representation allowed */
void
mpr_gen__div__invert_mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  if (w != x)
    {
      mpr_invert (w, y, m);
      mpr_mul (w, w, x, m);
    }
  else
    {
      mpr_t  t;
      TMP_DECL (marker);
      TMP_MARK (marker);
      MPR_TMP_INIT (t, m);
      mpr_invert (t, y, m);
      mpr_mul (w, x, t, m);
      TMP_FREE (marker);
    }
}

/* for unique representations */
int
mpr_gen__equal (mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  MPN_EQUAL_RETURN (RPTR(x), RPTR(y), MRSIZE(m));
}

/* for -1 uniquely as MNEGONE(m) */
int
mpr_gen__equal_negone (mpr_srcptr x, mpm_srcptr m)
{
  ASSERT (MNEGONE(m) != NULL);
  MPN_EQUAL_RETURN (RPTR(x), RPTR(MNEGONE(m)), MRSIZE(m));
}

/* for zero uniquely as all zeros */
int
mpr_gen__equal_zero (mpr_srcptr x, mpm_srcptr m)
{
  MPN_NONZERO_P_ACTION (RPTR(x), MRSIZE(m), return 0);
  return 1;
}

/* m odd and any congruent residue */
void
mpr_gen__gcd__odd (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;
  mp_ptr     txp, tmp;
  mp_size_t  xsize, msize;
  mp_limb_t  xlow;
  TMP_DECL (marker);

  ASSERT_MPN (RPTR(x), MRSIZE(m));
  ASSERT ((MPTR(m)[0] & 1) != 0);

  msize = MRSIZE(m);
  xsize = msize;
  xp = RPTR(x);

  /* strip high zero limbs from x */
  for (;;)
    {
      if (LIKELY (xp[xsize-1] != 0))
        break;
      xsize--;
      if (UNLIKELY (xsize == 0))
        {
          /* gcd(0,m) = m */
          SIZ(w) = msize;
          MPN_COPY (PTR(w), MPTR(m), msize);
          return;
        }
    }

  /* strip low zero limbs from x */
  for (;;)
    {
      xlow = xp[0];
      if (LIKELY (xlow != 0))
        break;
      xp++;
      xsize--;
      ASSERT (xsize > 0);
    }

  MPZ_REALLOC (w, xsize);

  TMP_MARK (marker);
  TMP_ALLOC_LIMBS_2 (txp,xsize, tmp,msize);

  /* temporary copy of x, and strip low zero bits */
  if ((xlow & 1) != 0)
    MPN_COPY (txp, xp, xsize);
  else
    {
      unsigned  shift;
      count_trailing_zeros (shift, xlow);
      mpn_rshift (txp, xp, xsize, shift);
      xsize -= (txp[xsize-1] == 0);
    }

  /* temporary copy of m */
  MPN_COPY (tmp, MPTR(m), msize);

  SIZ(w) = mpn_gcd (PTR(w), tmp, msize, txp, xsize);

  TMP_FREE (marker);
}

/* using mpr_get_z and expecting that routine won't need more than MRSIZE(m)
   limbs in z */
unsigned long
mpr_gen__get_ui__get_z (mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t      msize;
  unsigned long  w;
  mpz_t          z;
  TMP_DECL (marker);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  PTR(z) = TMP_ALLOC_LIMBS (msize);
  ALLOC(z) = msize;

  mpr_get_z (z, x, m);
  ASSERT (ALLOC(z) == msize);
  w = mpz_get_ui (z);

  TMP_FREE (marker);
  ASSERT_MPR_GET_UI (w, m);
  return w;
}

mpr_ptr
mpr_gen__new (mpm_srcptr m)
{
  mpr_ptr  r;
  r = (mpr_ptr) __GMP_ALLOCATE_FUNC_LIMBS (MRALLOC(m));
  ASSERT_CODE (MPN_FILL (RPTR(r), MRALLOC(m), DEADBEEF));
  return r;
}

/* x+m is used for the first operand, so as to get the inverse out as the
   first cofactor.  That cofactor can't be written to w directly, since
   mpn_gcdext clobbers an extra limb.  */
int
mpr_gen__invert (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  gsize, isize, xsize, msize;
  mp_srcptr  mp;
  mp_ptr     txp, tmp, gp, ip;
  mp_limb_t  c, glow;
  int        ret;
  TMP_DECL (marker);

  ASSERT_MPN (RPTR(x), MRSIZE(m));

  TMP_MARK (marker);
  msize = MRSIZE(m);

  TMP_ALLOC_LIMBS_4 (txp,msize+2, tmp,msize+1, gp,msize+1, ip,msize+2);

  /* txp = x + m */
  xsize = msize;
  mp = MPTR(m);
  c = mpn_add_n (txp, mp, RPTR(x), msize);
  txp[xsize] = c;
  xsize += (c != 0);

  /* tmp = m */
  MPN_COPY (tmp, mp, msize);

  gsize = mpn_gcdext (gp, ip,&isize, txp,xsize, tmp,msize);
  glow = gp[0];
  if (gsize != 1 || glow != 1)
    {
      ret = 0;
      goto done;
    }

  TRACE (mpn_trace ("  i", ip, isize));
  ASSERT (-msize <= isize && isize <= msize);
  if (isize >= 0)
    MPN_COPY_ZERO_EXTEND (RPTR(w), msize, ip, ABS(isize));
  else
    ASSERT_NOCARRY (mpn_sub (RPTR(w), mp, msize, ip, -isize));
  TRACE (mpn_trace ("  iw", RPTR(w), MRSIZE(m)));

  ASSERT_MPR_MOD (w, m);
  ret = 1;

 done:
  TMP_FREE (marker);
  return ret;
}

/* any congruent x */
int
mpr_gen__jacobi (mpr_srcptr x, mpm_srcptr m)
{
  mpz_t      xz;
  mp_srcptr  xp;
  mp_size_t  xsize;

  ASSERT_MPN (RPTR(x), MRSIZE(m));

  if ((MPTR(m)[0] & 1) == 0)
    DIVIDE_BY_ZERO;  /* not well-defined for m even */

  xp = RPTR(x);
  xsize = MRSIZE(m);
  MPN_NORMALIZE (xp, xsize);

  PTR(xz) = (mp_ptr) xp;
  SIZ(xz) = xsize;
  ASSERT_CODE (ALLOC(xz) = MRSIZE(m));

  return mpz_jacobi (xz, MZ(m));
}

void
mpr_gen__set (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  MPN_COPY (RPTR(w), RPTR(x), MRSIZE(m));
  ASSERT_CODE (MPN_FILL_EXTEND (RPTR(w), MRSIZE(m), MRALLOC(m), 0xDEADBEEF));
}

/* for ulong fitting in MRSIZE(m) limbs */
void
mpr_gen__set_ui__fits (mpr_ptr w, unsigned long x, mpm_srcptr m)
{
  mp_ptr  wp;
  wp = RPTR(w);
#if BITS_PER_ULONG <= GMP_NUMB_BITS
  ASSERT (MRSIZE(m) >= 1);
  wp[0] = x;
  MPN_ZERO_EXTEND (wp, (mp_size_t) 1, MRSIZE(m));
#else
  ASSERT (MRSIZE(m) >= 2);
  wp[0] = x & GMP_NUMB_MASK;
  wp[1] = x >> GMP_NUMB_BITS;
  MPN_ZERO_EXTEND (wp, (mp_size_t) 2, MRSIZE(m));
#endif
}

void
mpr_gen__swap (mpr_ptr x, mpr_ptr y, mpm_srcptr m)
{
  MPN_SWAP (RPTR(x), RPTR(y), MRSIZE(m));
}

