/* Main library mpm functions.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "tests.h"

#include "mpm.h"
#include "mpm-impl.h"


/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x)


int
mpr_set_str (mpr_ptr r, const char *str, int base, mpm_srcptr m)
{
  mpz_t  z;
  int    ret;
  ret = mpz_init_set_str (z, str, base);
  if (ret == 0)
    mpr_set_z (r, z, m);
  mpz_clear (z);
  return ret;
}


void
mpr_pow_ui (mpr_ptr r, mpr_srcptr b, unsigned long e, mpm_srcptr m)
{
  MPZ_FAKE_UI_VAR (ez, e);
  mpr_pow (r, b, ez, m);
}

void
mpr_ui_pow_ui (mpr_ptr r, unsigned long b, unsigned long e, mpm_srcptr m)
{
  MPZ_FAKE_UI_VAR (ez, e);
  mpr_set_ui (r, b, m);
  mpr_pow (r, r, ez, m);
}


void
mpm_swap (mpm_ptr x, mpz_ptr y)
{
  mpm_t  t;
  memcpy (t, x, sizeof (mpm_t));
  memcpy (x, y, sizeof (mpm_t));
  memcpy (y, t, sizeof (mpm_t));
}

void
mpm_clear (mpm_ptr m)
{
  if (m->_mpm_clear)
    (*(m)->_mpm_clear) (m);
  mpz_clear (MZ(m));
}

void
mpm_init_ui (mpm_ptr m, unsigned long modulus)
{
  MPZ_FAKE_UI_VAR (mz, modulus);
  mpm_init (m, mz);
}

int
mpm_init_str (mpm_ptr m, const char *str, int base)
{
  mpz_t  z;
  int    ret;

  ret = mpz_init_set_str (z, str, base);
  mpm_init (m, z);
  mpz_clear (z);

  return ret;
}

/* Division if even, otherwise redc or redc1 according to size. */
void
mpm_init (mpm_ptr m, mpz_srcptr mz)
{
  mp_size_t  msize;

  msize = SIZ(mz);
  if (msize == 0 || (PTR(mz)[0] & 1) == 0)
    mpm_init_div (m, mz);
  else
    mpm_init_redc (m, mz);
}


/* Strong probable prime test. */
int
mpr_sprp_p (mpr_srcptr base, mpm_srcptr m)
{
  mpr_t          p;
  int            ret;
  unsigned long  s;

  TRACE (printf ("mpr_sprp_p\n"));
  TRACE (mpr_trace ("  base", base, m));
  if (mpz_even_p (mpm_modref(m)))
    return 0;


  /* find s satisfying m-1 = e*2^s, where e is odd */
  s = mpz_scan1 (mpm_modref(m), 1L);
  if (s == (unsigned long) -1)
    return 0;  /* m==1, not prime */

  mpr_init (p, m);

  {
    mpz_t e;
    mpz_init (e);
    mpz_fdiv_q_2exp (e, mpm_modref(m), s);
    TRACE (printf ("  s=%lu ", s); mpz_trace ("e", e));

    mpr_pow (p, base, e, m);
    TRACE (mpr_trace ("  p", p, m));
    mpz_clear (e);
  }

  if (mpr_equal_one (p, m))
    {
    yes:
      ret = 1;
      goto done;
    }

  for (;;)
    {
      if (mpr_equal_negone (p, m))
        goto yes;
      s--;
      if (s == 0)
        break;
      mpr_mul (p, p, p, m);
      TRACE (printf ("  s=%lu ", s); mpr_trace ("p", p, m));
    }
  ret = 0;

 done:
  mpr_clear (p, m);
  return ret;
}
