/* One-limb modular arithmetic.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "mpm.h"
#include "mpm-impl.h"
#include "redc1.h"


/* The modulus is a single limb, limited to nail size.
   A residue is a single limb in the range 0 <= r < modulus.

   This special case code is probably not worth bothering with.
   mpr_one__pow is about the only function where the call overhead doesn't
   swamp the operation.  Better use redc1.h.  */


#define MLIMB(m)   ((m)->_mp_norm_rshift_mask)

/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x) 

#if WANT_ASSERT
#define ASSERT_MPR_ONE(r,m)             \
  do {                                  \
    mp_limb_t  __mlimb = MLIMB(m);      \
    ASSERT (RPTR(r)[0] < __mlimb);      \
  } while (0)
#else
#define ASSERT_MPR_ONE(r,m)
#endif


void mpr_one__mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m);


static void
mpr_one__add (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  ASSERT_MPR_ONE (x, m);
  ADDMOD_1 (RPTR(w)[0], RPTR(x)[0], RPTR(y)[0], MLIMB(m));
  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__add_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_limb_t  mlimb;
  ASSERT_MPR_ONE (x, m);
  mlimb = MLIMB(m);
  if (UNLIKELY (y >= mlimb))
    y %= mlimb;
  ADDMOD_1 (RPTR(w)[0], RPTR(x)[0], (mp_limb_t) y, mlimb);
  ASSERT_MPR_ONE (w, m);
}

/* ENHANCE-ME: Maybe y>GMP_NUMB_MAX can be handled more efficiently. */
static void
mpr_one__addmul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_limb_t  hi, lo, dummy_q, mlimb, plimb;
  mp_ptr     wp;

  ASSERT_MPR_ONE (x, m);

  mlimb = MLIMB(m);
  if (UNLIKELY (y > GMP_NUMB_MAX))
    y %= mlimb;

  umul_ppmm (hi, lo, RPTR(x)[0], y);
  udiv_qrnnd_unnorm (dummy_q, plimb, hi,lo, mlimb);
  wp = RPTR(w);
  ADDMOD_1 (wp[0], wp[0], plimb, mlimb);

  ASSERT_MPR_ONE (w, m);
}

static int
mpr_one__equal_negone (mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_ONE (x, m);
  return (RPTR(x)[0] == MLIMB(m)-1);
}

static int
mpr_one__equal_one (mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_ONE (x, m);
  return (RPTR(x)[0] == 1);
}

static int
mpr_one__equal_ui (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_limb_t  mlimb;

  ASSERT_MPR_ONE (x, m);

  mlimb = MLIMB(m);
  if (UNLIKELY (y >= mlimb))
    y %= mlimb;

  return (RPTR(x)[0] == y);
}

static void
mpr_one__gcd (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_limb_t  xlimb, wlimb;

  ASSERT_MPR_ONE (x, m);

  xlimb = RPTR(x)[0];
  wlimb = MLIMB(m);   /* gcd(0,m)==m */
  if (LIKELY (xlimb != 0))
    wlimb = mpn_gcd_1 (&MLIMB(m), (mp_size_t) 1, xlimb);

  PTR(w)[0] = wlimb;
  SIZ(w) = (wlimb != 0);

  ASSERT_MPZ (w);
}

static unsigned long
mpr_one__get_ui (mpr_srcptr x, mpm_srcptr m)
{
  unsigned long  w;
  ASSERT_MPR_ONE (x, m);
  w = RPTR(x)[0];
  ASSERT_MPR_GET_UI (w, m);
  return w;
}

static void
mpr_one__get_z (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_limb_t  xlimb;
  ASSERT_MPR_ONE (x, m);

  xlimb = RPTR(x)[0];
  PTR(w)[0] = xlimb;
  SIZ(w) = (xlimb != 0);

  ASSERT_MPR_GET_Z (w, m);
}

void
mpr_one__mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_limb_t  hi, lo, dummy_q;

  ASSERT_MPR_ONE (x, m);
  ASSERT_MPR_ONE (y, m);

  umul_ppmm (hi, lo, RPTR(x)[0], RPTR(y)[0]);
  udiv_qrnnd_unnorm (dummy_q, RPTR(w)[0], hi,lo, MLIMB(m));

  ASSERT_MPR_ONE (w, m);
}

/* ENHANCE-ME: Maybe y>GMP_NUMB_MAX can be handled more efficiently. */
static void
mpr_one__mul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_limb_t  hi, lo, dummy_q, mlimb;

  ASSERT_MPR_ONE (x, m);

  mlimb = MLIMB(m);
  if (UNLIKELY (y > GMP_NUMB_MAX))
    y %= mlimb;

  umul_ppmm (hi, lo, RPTR(x)[0], y);
  udiv_qrnnd_unnorm (dummy_q, RPTR(w)[0], hi,lo, mlimb);

  ASSERT_MPR_ONE (w, m);
}

void
mpr_one__mul_si (mpr_ptr w, mpr_srcptr x, long y_signed, mpm_srcptr m)
{
  mp_limb_t  hi, lo, dummy_q, mlimb, wlimb;
  mp_limb_t  y;

  ASSERT_MPR_ONE (x, m);

  mlimb = MLIMB(m);
  y = ABS (y_signed);
  if (UNLIKELY (y > GMP_NUMB_MAX))
    y %= mlimb;

  umul_ppmm (hi, lo, RPTR(x)[0], y);
  udiv_qrnnd_unnorm (dummy_q, wlimb, hi,lo, mlimb);
  if (y_signed < 0)
    NEGMOD_1 (wlimb, wlimb, mlimb);
  RPTR(w)[0] = wlimb;

  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__neg (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_ONE (x, m);
  NEGMOD_1 (RPTR(w)[0], RPTR(x)[0], MLIMB(m));
  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__pow (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  esize, i;
  mp_srcptr  ep;
  mp_limb_t  blimb, elimb, mlimb, rlimb, hi, lo, mask, dummy_q;
  int        lz;

  ASSERT_MPR_ONE (b, m);
  ASSERT_MPZ (e);

  esize = SIZ(e);
  if (UNLIKELY (esize == 0))
    {
      /* b^0 == 1 */
      rlimb = 1;
      goto done;
    }

  blimb = RPTR(b)[0];
  if (UNLIKELY (esize < 0))
    {
      esize = -esize;
      if (! mpr_gen__invert (r, b, m))
        DIVIDE_BY_ZERO;
      blimb = RPTR(r)[0];
    }

  rlimb = blimb;  /* start b^1 */
  mlimb = MLIMB(m);

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (lz, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - lz);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }

      /* r^2 mod m */
      umul_ppmm (hi, lo, rlimb, rlimb);
      udiv_qrnnd_unnorm (dummy_q, rlimb, hi,lo, mlimb);

      if ((elimb & mask) != 0)
        {
          /* r*b mod m */
          umul_ppmm (hi, lo, rlimb, blimb);
          udiv_qrnnd_unnorm (dummy_q, rlimb, hi,lo, mlimb);
        }
    }

 done:
  RPTR(r)[0] = rlimb;
}

static void
mpr_one__set_si (mpr_ptr w, long x, mpm_srcptr m)
{
  mp_limb_t      mlimb, wlimb;
  unsigned long  xu;

  xu = ABS(x);
  mlimb = MLIMB(m);
  if (UNLIKELY (xu >= mlimb))
    xu %= mlimb;
  wlimb = xu;
  if (x < 0)
    NEGMOD_1 (wlimb, wlimb, mlimb);
  RPTR(w)[0] = wlimb;

  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__set_ui (mpr_ptr w, unsigned long x, mpm_srcptr m)
{
  mp_limb_t  mlimb;

  mlimb = MLIMB(m);
  if (UNLIKELY (x >= mlimb))
    x %= mlimb;
  RPTR(w)[0] = x;

  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__set_z (mpr_ptr w, mpz_srcptr x, mpm_srcptr m)
{
  mp_limb_t  wlimb, mlimb;
  mp_size_t  xsize;

  xsize = SIZ(x);
  mlimb = MLIMB(m);
  wlimb = mpn_mod_1 (PTR(x), ABS(xsize), mlimb);
  if (xsize < 0)
    NEGMOD_1 (wlimb, wlimb, mlimb);
  RPTR(w)[0] = wlimb;

  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__sub (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  ASSERT_MPR_ONE (x, m);
  SUBMOD_1 (RPTR(w)[0], RPTR(x)[0], RPTR(y)[0], MLIMB(m));
  ASSERT_MPR_ONE (w, m);
}

static void
mpr_one__sub_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_limb_t  mlimb;

  ASSERT_MPR_ONE (x, m);

  mlimb = MLIMB(m);
  if (UNLIKELY (y >= mlimb))
    y %= mlimb;
  SUBMOD_1 (RPTR(w)[0], RPTR(x)[0], (mp_limb_t) y, MLIMB(m));
  ASSERT_MPR_ONE (w, m);
}

/* ENHANCE-ME: Maybe y>GMP_NUMB_MAX can be handled more efficiently. */
static void
mpr_one__submul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_limb_t  hi, lo, dummy_q, mlimb, plimb;
  mp_ptr     wp;

  ASSERT_MPR_ONE (x, m);

  mlimb = MLIMB(m);
  if (UNLIKELY (y > GMP_NUMB_MAX))
    y %= mlimb;

  umul_ppmm (hi, lo, RPTR(x)[0], y);
  udiv_qrnnd_unnorm (dummy_q, plimb, hi,lo, mlimb);
  wp = RPTR(w);
  SUBMOD_1 (wp[0], wp[0], plimb, mlimb);

  ASSERT_MPR_ONE (w, m);
}


void
mpm_init_one (mpm_ptr m, mp_limb_t mlimb)
{
  mp_ptr     mp;

  if (mlimb < 2)
    DIVIDE_BY_ZERO;

  ASSERT_CODE (memset (m, '\0', sizeof (mpm_t)));

  MLIMB(m) = mlimb;
  MRSIZE(m) = 1;
  MRALLOC(m) = 1;

  ALLOC(MZ(m)) = 1;
  SIZ(MZ(m)) = 1;
  mp = __GMP_ALLOCATE_FUNC_LIMBS (1);
  MPTR(m) = mp;
  mp[0] = mlimb;

  m->_mp_smallinv[0] = mlimb/2 + (mlimb&1);
  MDIV2CEIL(m) = &m->_mp_smallinv[0];

#if 0
  /* not needed */
  m->_mp_smallinv[1] = mlimb-1;
  MNEGONE(m) = &m->_mp_smallinv[1];
#endif

  m->_mpm_clear = NULL;

  m->_mpr_add          = mpr_one__add;
  m->_mpr_add_ui       = mpr_one__add_ui;
  m->_mpr_addmul_ui    = mpr_one__addmul_ui;
  m->_mpr_clear        = mpr_gen__clear;
  m->_mpr_div          = mpr_gen__div__invert_mul;
  m->_mpr_div_2exp     = mpr_mod__div_2exp;
  m->_mpr_equal        = mpr_gen__equal;
  m->_mpr_equal_one    = mpr_one__equal_one;
  m->_mpr_equal_negone = mpr_one__equal_negone;
  m->_mpr_equal_ui     = mpr_one__equal_ui;
  m->_mpr_equal_zero   = mpr_gen__equal_zero;
  m->_mpr_gcd          = mpr_one__gcd;
  m->_mpr_get_ui       = mpr_one__get_ui;
  m->_mpr_get_z        = mpr_one__get_z;
  m->_mpr_invert       = mpr_gen__invert;
  m->_mpr_jacobi       = mpr_gen__jacobi;
  m->_mpr_mul          = mpr_one__mul;
  m->_mpr_mul_2exp     = mpr_mod__mul_2exp;
  m->_mpr_mul_si       = mpr_one__mul_si;
  m->_mpr_mul_ui       = mpr_one__mul_ui;
  m->_mpr_neg          = mpr_one__neg;
  m->_mpr_new          = mpr_gen__new;
  m->_mpr_pow          = mpr_one__pow;
  m->_mpr_set          = mpr_gen__set;
  m->_mpr_set_z        = mpr_one__set_z;
  m->_mpr_set_si       = mpr_one__set_si;
  m->_mpr_set_ui       = mpr_one__set_ui;
  m->_mpr_sub          = mpr_one__sub;
  m->_mpr_sub_ui       = mpr_one__sub_ui;
  m->_mpr_submul_ui    = mpr_one__submul_ui;
  m->_mpr_swap         = mpr_gen__swap;
  m->_mpr_urandomm     = mpr_mod__urandomm;
}
